from os.path import exists
from os.path import getmtime

from baelfire.dependencies.dependency import Dependency


class FileDependency(Dependency):
    """
    Base file dependency.
    """

    def __init__(self, name=None, raw_path=None):
        super(FileDependency, self).__init__()
        self.source_name = name
        self.raw_path = raw_path

    def phase_data(self):
        super(FileDependency, self).phase_data()
        self.myreport['filename'] = self.path

    @property
    def path(self):
        return self.raw_path or self.paths.get(self.source_name)


class FileChanged(FileDependency):
    """
    Trigger build if dependency file was changed.
    """

    def should_build(self):
        try:
            output = getmtime(self.parent.output)
        except OSError:
            return True
        try:
            source = getmtime(self.path)
        except OSError:
            return True

        return output < source


class FileExists(FileDependency):
    """
    Trigger build if dependency file exists.
    """

    def should_build(self):
        return exists(self.path)


class FileDoesNotExists(FileDependency):
    """
    Trigger build if dependency file does not exists.
    """

    def should_build(self):
        return not exists(self.path)



from tempfile import NamedTemporaryFile

from mock import patch
from pytest import raises

from baelfire.filedict import FileDict


class TestFileDict(object):

    @property
    def path(self):
        return NamedTemporaryFile().name

    def obj(self):
        return FileDict(self.path)

    def test_saving(self):
        data = self.obj()
        data['first'] = 10
        data['second'] = 'xxx'
        data.save()

        loader = FileDict(data.path)
        loader.load()

        assert dict(data) == dict(loader)

    def test_loading(self):
        data = self.obj()
        data['first'] = 10
        data['second'] = 'xxx'
        data.save()

        data['third'] = 'lel'
        data.load()

        assert dict(data) == {
            'first': 10,
            'second': 'xxx',
        }

    def test_loading_on_error(self):
        data = self.obj()
        with raises(IOError):
            data.load()
        assert dict(data) == {}

    def test_loading_with_ignore_error(self):
        assert dict(self.obj().load(True)) == {}

    def test_ensure_key_not_exists(self):
        obj = self.obj()
        patcher = patch.object(obj, 'read_from_stdin')

        with patcher as mock:
            obj.ensure_key_exists('key', 'description')
            mock.assert_called_once_with('key', 'description')

    def test_ensure_key_exists(self):
        obj = self.obj()
        patcher = patch.object(obj, 'read_from_stdin')
        obj['key'] = 'x'

        with patcher as mock:
            obj.ensure_key_exists('key', 'description')
            assert mock.called is False

    def test_read_from_stdin(self):
        obj = self.obj()
        patcher = patch('baelfire.filedict.input')

        with patcher as mock:
            obj.read_from_stdin('key', 'description')
            mock.assert_called_once_with('description: ')
            assert obj['key'] == mock.return_value

    def test_read_from_stdin_without_description(self):
        obj = self.obj()
        patcher = patch('baelfire.filedict.input')

        with patcher as mock:
            obj.read_from_stdin('key')
            mock.assert_called_once_with('')
            assert obj['key'] == mock.return_value



import signal

from subprocess import Popen

from baelfire.error import CommandAborted
from baelfire.error import CommandError
from baelfire.task.task import Task


class SubprocessTask(Task):
    """
    Task which will run external programs. This task will manage sending
    proccess signals.
    """

    def phase_init(self):
        super(SubprocessTask, self).phase_init()
        self.spp = None
        self.myreport['aborted'] = False
        self.myreport['signal'] = None

    def send_signal(self, signal):
        try:
            if self.spp.poll() is None:
                self.spp.send_signal(signal)
        except OSError:
            pass

    def on_signal(self, signum, frame):
        self.myreport['aborted'] = True
        self.myreport['signal'] = signum
        self.spp and self.send_signal(signum)

    def popen(self, *args, **kwargs):
        """
        Run subprocess.Popen whitin the task.
        """
        args, kwargs = self._set_default_args(args, kwargs)
        self._init_signals()
        self._popen(*args, **kwargs)
        self._post_popen()

    def _set_default_args(self, args, kwargs):
        kwargs.setdefault('shell', True)
        return args, kwargs

    def _init_signals(self):
        for _signal in [
            signal.SIGABRT,
            signal.SIGFPE,
            signal.SIGILL,
            signal.SIGINT,
            signal.SIGSEGV,
            signal.SIGTERM,
        ]:
            signal.signal(_signal, self.on_signal)

    def _popen(self, *args, **kwargs):
        self.spp = Popen(*args, **kwargs)
        self.spp.wait()

    def _post_popen(self):
        if self.myreport['aborted']:
            raise CommandAborted()
        if self.spp.returncode is not 0:
            raise CommandError(self.spp.returncode)



class Dependency(object):

    @property
    def name(self):
        """
        Return name of this task.
        """
        return self.__module__ + "." + self.__class__.__name__

    def __init__(self):
        self.task = None
        self.parent = None
        self.myreport = {
            'name': self.name,
        }

    def phase_init(self):
        self.myreport['phase_validation'] = False
        self.myreport['builded'] = False
        self.myreport['success'] = False
        self.myreport['should_build'] = None
        self.myreport['index'] = self.parent.get_index()

    def phase_data(self):
        pass

    def phase_validation(self):
        self.myreport['phase_validation'] = True
        result = self.should_build()
        self.myreport['should_build'] = result
        self.parent.logger.debug(
            'Dependency %(dependency)s result: %(result)r' % {
                'dependency': self.name,
                'result': result,
            }
        )
        return result

    def phase_build(self):
        self.myreport['builded'] = True
        self.build()
        self.myreport['success'] = True

    def set_parent(self, parent):
        self.parent = parent

    @property
    def settings(self):
        return self.parent.settings

    @property
    def paths(self):
        return self.parent.paths

    def build(self):
        pass


class AlwaysTrue(Dependency):
    """
    Always rebuild the task.
    """

    def should_build(self):
        return True


class AlwaysFalse(Dependency):
    """
    This dependency will always return "do not rebuild". It is for testing
    purpose.
    """

    def should_build(self):
        return False



from mock import MagicMock

from baelfire.dependencies.dependency import AlwaysTrue
from baelfire.dependencies.dependency import Dependency
from baelfire.dependencies.dependency import AlwaysFalse


class TestAlwaysTrue(object):

    def test_should_build(self):
        """
        AlwaysTrue.should_rebuild should always return True.
        """
        assert AlwaysTrue().should_build() is True


class TestAlwaysFalse(object):

    def test_should_build(self):
        """
        AlwaysFalse.should_rebuild should always return False.
        """
        assert AlwaysFalse().should_build() is False


class TestDependency(object):

    def test_settings(self):
        """
        .settings should return parent settings.
        """
        parent = MagicMock()
        dependency = Dependency()
        dependency.set_parent(parent)

        assert dependency.settings is parent.settings



import os

from baelfire.dependencies.dependency import Dependency


class PidIsRunning(Dependency):
    """
    Trigger build if pid is running.
    """

    def __init__(self, pid=None, pid_file_name=None, pid_file_path=None):
        super(PidIsRunning, self).__init__()
        self.pid = pid
        self.pid_file_name = pid_file_name
        self.pid_file_path = pid_file_path

    def get_pid(self):
        if self.pid:
            return self.pid
        elif self.pid_file_name:
            return self._get_pid_from_file(self.paths[self.pid_file_name])
        else:
            return self._get_pid_from_file(self.pid_file_path)

    def _get_pid_from_file(self, path):
        with open(path, 'r') as pidfile:
            return int(pidfile.read())

    def is_running(self):
        try:
            os.kill(self.get_pid(), 0)
            return True
        except OSError:
            return False

    def should_build(self):
        return self.is_running()


class PidIsNotRunning(PidIsRunning):
    """
    Trigger build if pid is not running.
    """

    def should_build(self):
        return not self.is_running()



from baelfire.application.commands.graph.visualizators.chooser import VisualizatorChooser
from baelfire.application.commands.graph.visualizators.dependency import DependencyVisualizatorChooser
from baelfire.application.commands.graph.visualizators.link import LinkVisualizatorChooser


class TaskVisualization(object):
    PREFIX = '\t'

    class Configure(object):
        dependency = DependencyVisualizatorChooser
        link = LinkVisualizatorChooser

    class Templates(object):
        task = '"%(url)s"[label="%(name)s",fillcolor=%(fill)s,style=filled];\n'
        always_run = '"%(url)s"[label="%(name)s",fillcolor=%(fill)s,style=filled,shape=octagon];\n'

    def __init__(self, url, report):
        self.url = url
        self.report = report

    def task(self):
        if self.report['needtorun']:
            if self.report['success']:
                fill = 'green'
            else:
                fill = 'red'
        else:
            fill = 'white'
        return {
            'url': self.url,
            'name': self.url.split('.')[-1],
            'fill': fill,
        }

    def render(self):
        self.rendered = ''
        self.render_task()
        for dependency in self.report['dependencies']:
            vdependency = self.Configure.dependency().choose(
                dependency['name'],
                dependency,
                self,
            )
            vlink = self.Configure.link().choose(
                self.url,
                self.report,
                vdependency.full_url,
                dependency,
            )
            self.rendered += vdependency.render()
            self.rendered += vlink.render()

        return self.rendered

    def render_task(self):
        always_rebuild = 'baelfire.dependencies.dependency.AlwaysTrue'
        if always_rebuild in self.get_dependencies_names():
            self.rendered += (
                self.PREFIX + (self.Templates.always_run % self.task())
            )
        else:
            self.rendered += self.PREFIX + (self.Templates.task % self.task())

    def get_dependencies_names(self):
        return [item['name'] for item in self.report['dependencies']]


class TaskVisualizatorChooser(VisualizatorChooser):
    DEFAULT = TaskVisualization

    def generate_visualizators(self):
        pass



from baelfire.dependencies.dependency import AlwaysTrue
from baelfire.task import SubprocessTask
from baelfire.core import Core


class ScreenCore(Core):
    """
    Core with 'exe:screen' setting, which gives path to the screen application.
    """

    def phase_settings(self):
        super(ScreenCore, self).phase_settings()
        self.paths.set('exe:screen', ['usr', 'bin', 'screen'], is_root=True)


class ScreenTask(SubprocessTask):
    """
    Start command in detached screen.
    In order to use ScreenTask, your application's core needs to inherit from ScreenCore.
    """

    screen_name = None

    def _screen_run(self, *args, **kwargs):
        screen_name = 'S %s' % (self.screen_name,) if self.screen_name else ''
        args[0][0] = (
            '%s -dm%s ' % (self.paths.get('exe:screen'), screen_name) + args[0][0]
        )
        return self.popen(*args, **kwargs)


class AttachScreenTask(SubprocessTask):
    """
    Attach task from screen. Run it first if needed.

    In order to use AttachScreenTask, your application's core needs to inherit from ScreenCore and you need to set
    `detached_task` property, which should be the ScreenTask like task.
    """
    detached_task = None

    def create_dependecies(self):
        assert self.detached_task
        self._detached_task = self.detached_task()

        self.run_before(self._detached_task)
        self.build_if(AlwaysTrue())

    def _screen_attach(self):
        cmd = '%s -r %s' % (self.paths.get('exe:screen'),
                            self._detached_task.screen_name)
        return self.popen([cmd])

    def build(self):
        self._screen_attach()



from baelfire.task.task import Task

from baelfire.task.file import FileTask
from baelfire.task.process import SubprocessTask
from baelfire.task.template import TemplateTask

__all__ = [
    'Task',
    'FileTask',
    'TemplateTask',
    'SubprocessTask',
]



from subprocess import PIPE
from subprocess import Popen
import re

from baelfire.dependencies.dependency import Dependency


class ScreenIsNotRunning(Dependency):
    regex = re.compile(r'^\s*([^\.]*)\.([^\s]+)\s*\((\w+)\)$', re.UNICODE)

    def __init__(self, screen_name=None):
        super(ScreenIsNotRunning, self).__init__()
        self._screen_name = screen_name

    def _list_screen(self):
        cmd = self.paths.get('exe:screen') + ' -list'
        spp = Popen([cmd], shell=True, stdout=PIPE)
        spp.wait()
        return spp.stdout

    def get_screens(self):
        for line in self._list_screen().readlines():
            line = line.decode()
            try:
                data = self.regex.findall(str(line))[0]
                yield {
                    'pid': int(data[0]),
                    'name': data[1],
                    'status': data[2],
                }
            except IndexError:
                pass

    def should_build(self):
        for screen in self.get_screens():
            if screen['name'] == self.screen_name:
                return False
        return True

    @property
    def screen_name(self):
        return self._screen_name or self.parent.screen_name



from baelfire.dependencies.file import FileDoesNotExists
from baelfire.task.task import Task


class FileTask(Task):
    """
    Task which operates on files.

    Need to implement ``FileTask.output_name`` parameter.
    """

    @property
    def output(self):
        return self.paths.get(self.output_name)

    def create_dependecies(self):
        try:
            self.build_if(FileDoesNotExists(self.output_name))
        except AttributeError:
            self.build_if(FileDoesNotExists(raw_path=self.output))



def parrented(method):
    def wrapper(self, *args, **kwargs):
        if self.parent:
            method_name = method.__name__
            parent_method = getattr(self.parent, method_name)
            if hasattr(parent_method, '__call__'):
                return parent_method(self, *args, **kwargs)
            else:
                return parent_method
        else:
            return method(self, *args, **kwargs)
    return wrapper



from tempfile import NamedTemporaryFile

from baelfire.task import FileTask
from os import unlink
from os.path import exists


class ExampleFileTask(FileTask):
    output_name = 'file'

    def phase_settings(self):
        self.paths.set('file', ['/tmp', 'file.txt'])

    def build(self):
        open(self.output, 'w').close()


class ExampleSecondFileTask(FileTask):

    def build(self):
        open(self.output, 'w').close()


class TestFileTask(object):

    def test_normal(self):
        """
        FileTask should rebuild when file does not exists.
        """
        task = ExampleFileTask()
        task.phase_init()
        task.phase_settings()
        try:
            unlink(task.output)
        except:
            pass

        task.run()

        assert exists(task.output)
        assert task.report == {
            'baelfire.task.tests.test_file.ExampleFileTask': {
                'dependencies': [
                    {
                        'builded': True,
                        'filename': '/tmp/file.txt',
                        'index': 0,
                        'phase_validation': True,
                        'should_build': True,
                        'success': True,
                        'name': 'baelfire.dependencies.file.FileDoesNotExists',
                    },
                ],
                'needtorun': True,
                'runned': True,
                'success': True,
            },
            'last_index': 1,
        }

        task.run()
        assert task.report == {
            'baelfire.task.tests.test_file.ExampleFileTask': {
                'dependencies': [
                    {
                        'builded': True,
                        'filename': '/tmp/file.txt',
                        'index': 0,
                        'phase_validation': True,
                        'should_build': False,
                        'success': True,
                        'name': 'baelfire.dependencies.file.FileDoesNotExists',
                    },
                ],
                'needtorun': False,
                'runned': False,
                'success': False,
            },
            'last_index': 1,
        }

    def test_raw_path(self):
        ExampleSecondFileTask.output = NamedTemporaryFile().name
        task = ExampleSecondFileTask()
        task.run()

        assert task.report == {
            'baelfire.task.tests.test_file.ExampleSecondFileTask': {
                'dependencies': [
                    {
                        'builded': True,
                        'filename': ExampleSecondFileTask.output,
                        'index': 0,
                        'phase_validation': True,
                        'should_build': True,
                        'success': True,
                        'name': 'baelfire.dependencies.file.FileDoesNotExists',
                    },
                ],
                'needtorun': True,
                'runned': True,
                'success': True,
            },
            'last_index': 1,
        }



from mock import MagicMock
from mock import patch
from pytest import fixture
from pytest import raises
from pytest import yield_fixture
from yaml import load

from baelfire.core import Core
from baelfire.dependencies import AlwaysTrue
from baelfire.dependencies.dependency import AlwaysFalse
from baelfire.dependencies.task import TaskRebuilded
from baelfire.task import Task


class ExampleCore(Core):

    def phase_settings(self):
        super(ExampleCore, self).phase_settings()
        self.settings['child'] = '2'
        self.paths.set('child', 'child.txt')

        self.settings['parent'] = '1'
        self.paths.set('parent', 'parent.txt')


class ExampleTask(Task):

    def create_dependecies(self):
        pass

    def build(self):
        pass


class ExampleFailingTask(Task):

    def create_dependecies(self):
        self.build_if(AlwaysTrue())

    def build(self):
        raise RuntimeError('failing')


class ExampleChild(ExampleTask):

    def build(self):
        self._data = {
            'settings': {
                'child': self.settings['child'],
                'parent': self.settings['parent']
            },
            'paths': {
                'child': self.paths.get('child'),
                'parent': self.paths.get('parent'),
            },
        }


class ExampleParent(ExampleTask):

    def build(self):
        self._data = {
            'settings': {
                'child': self.settings['child'],
                'parent': self.settings['parent']
            },
            'paths': {
                'child': self.paths.get('child'),
                'parent': self.paths.get('parent'),
            },
        }


class TestTask(object):

    @fixture
    def task(self):
        return ExampleTask()

    @yield_fixture
    def mrun_task(self):
        with patch('baelfire.task.task.RunTask') as mock:
            yield mock

    def test_simple_flow(self, task):
        """
        Sanity check. This test simple run without dependecies or build.
        """
        task.run()

        assert task.report == {
            'baelfire.task.tests.test_tasks.ExampleTask': {
                'dependencies': [],
                'needtorun': False,
                'runned': False,
                'success': False,
            },
            'last_index': 0,
        }

    def test_dependency_run(self, task):
        """
        AlwaysTrue should indicate rebuilding task every time. AlwaysFalse
        should not indicate rebuilding at any time. This test tests report for
        dependencies.
        """
        task.build_if(AlwaysTrue())
        task.build_if(AlwaysFalse())

        task.run()

        assert task.report == {
            'baelfire.task.tests.test_tasks.ExampleTask': {
                'dependencies': [
                    {
                        'builded': True,
                        'index': 0,
                        'should_build': True,
                        'phase_validation': True,
                        'success': True,
                        'name': 'baelfire.dependencies.dependency.AlwaysTrue',
                    },
                    {
                        'builded': True,
                        'index': 1,
                        'should_build': False,
                        'phase_validation': True,
                        'success': True,
                        'name': 'baelfire.dependencies.dependency.AlwaysFalse',
                    },
                ],
                'needtorun': True,
                'runned': True,
                'success': True,
            },
            'last_index': 2,
        }

    def test_exception(self, task):
        """
        .report should make success to false, when task building raise an
        error.
        """
        task = ExampleFailingTask()

        with raises(RuntimeError):
            task.run()

        assert task.report == {
            'baelfire.task.tests.test_tasks.ExampleFailingTask': {
                'dependencies': [
                    {
                        'builded': True,
                        'index': 0,
                        'should_build': True,
                        'phase_validation': True,
                        'success': True,
                        'name': 'baelfire.dependencies.dependency.AlwaysTrue',
                    },
                ],
                'needtorun': True,
                'runned': True,
                'success': False,
            },
            'last_index': 1,
        }

    def test_data_parrenting(self):
        """
        When task is a child of another, it should share this properties:
            - settings
            - paths
            - report
        """
        parent = ExampleParent(ExampleCore())
        child = ExampleChild()
        child.build_if(AlwaysTrue())
        parent.build_if(TaskRebuilded(child))

        parent.run()

        assert parent._data == child._data
        assert parent.report == {
            'baelfire.task.tests.test_tasks.ExampleChild': {
                'dependencies': [
                    {
                        'builded': True,
                        'index': 1,
                        'should_build': True,
                        'phase_validation': True,
                        'success': True,
                        'name': 'baelfire.dependencies.dependency.AlwaysTrue',
                    },
                ],
                'needtorun': True,
                'runned': True,
                'success': True,
            },
            'baelfire.task.tests.test_tasks.ExampleParent': {
                'dependencies': [
                    {
                        'builded': True,
                        'index': 0,
                        'phase_validation': True,
                        'should_build': True,
                        'success': True,
                        'task': 'baelfire.task.tests.test_tasks.ExampleChild',
                        'name': 'baelfire.dependencies.task.TaskRebuilded',
                    }
                ],
                'needtorun': True,
                'runned': True,
                'success': True,
            },
            'last_index': 2,
        }

    def test_saving(self):
        """
        .save_report should save report to a file with yaml
        """
        task = ExampleFailingTask()

        with raises(RuntimeError):
            task.run()

        task.save_report()

        data = load(open(task.paths.get('report'), 'r').read())

        assert data == {
            'baelfire.task.tests.test_tasks.ExampleFailingTask': {
                'dependencies': [
                    {
                        'builded': True,
                        'index': 0,
                        'should_build': True,
                        'phase_validation': True,
                        'success': True,
                        'name': 'baelfire.dependencies.dependency.AlwaysTrue',
                    },
                ],
                'needtorun': True,
                'runned': True,
                'success': False,
            },
            'last_index': 1,
        }

    def test_run_before(self, task, mrun_task):
        """
        .run_before should add RunTask dependency to the dependency list.
        """
        child_task = MagicMock()
        task.run_before(child_task)

        assert task._dependencies == [mrun_task.return_value]
        mrun_task.assert_called_once_with(child_task)
        mrun_task.return_value.set_parent.assert_called_once_with(task)



from mock import MagicMock
from mock import patch
from pytest import fixture
from pytest import raises
from pytest import yield_fixture

from baelfire.application.application import Application
from baelfire.application.application import run
from baelfire.error import TaskNotFoundError
from baelfire.task import Task


class ExampleTask(Task):

    def create_dependecies(self):
        pass

    def build(self):
        pass


class TestApplication(object):

    @fixture
    def app(self):
        return Application()

    @yield_fixture
    def parse_args(self, app):
        app.create_parser()
        with patch.object(app.parser, 'parse_args') as parse_args:
            yield parse_args

    @yield_fixture
    def import_task(self, app):
        with patch.object(app, 'import_task') as mock:
            yield mock

    def test_run_without_args(self, app, parse_args):
        with patch.object(app.parser, 'print_help') as print_help:
            args = MagicMock()
            args.task = None
            args.graph_file = False

            app.run_command_or_print_help(args)

            print_help.assert_called_once_with()

    def test_run_with_task(self, app, parse_args):
        args = MagicMock()
        args.task = (
            'baelfire.application.tests.test_application:ExampleTask'
        )
        args.graph = False
        with patch.object(ExampleTask, 'phase_build') as phase_build:
            with patch.object(ExampleTask, 'save_report') as save_report:
                app.run_command_or_print_help(args)

                save_report.assert_called_once_with()
                phase_build.assert_called_once_with()

    def test_run_with_bad_task(self, app, parse_args):
        args = MagicMock()
        args.task = (
            'baelfire.application.tests.test_application:BadTask'
        )
        with raises(TaskNotFoundError):
            app.run_command_or_print_help(args)

    def test_run_with_task_error(self, app, import_task, parse_args):
        task = import_task.return_value.return_value
        task.run.side_effect = RuntimeError

        args = MagicMock()
        with raises(RuntimeError):
            app.run_command_or_print_help(args)

        import_task.assert_called_once_with(args.task)

    def test_run(self):
        parser_patcher = patch.object(
            Application,
            'create_parser',
            autospec=True,
        )
        run_patcher = patch.object(Application, 'run_command_or_print_help')
        with parser_patcher as create_parser:
            args = MagicMock()
            args.log_level = 'info'

            def side_effect(self):
                self.parser = MagicMock()
                self.parser.parse_args.return_value = args
            create_parser.side_effect = side_effect
            with run_patcher as run_command_or_print_help:
                run()
                assert create_parser.call_count == 1
                run_command_or_print_help.assert_called_once_with(args)

    def test_task_with_grap(self, app, parse_args):
        args = MagicMock()
        args.task = (
            'baelfire.application.tests.test_application:ExampleTask'
        )
        with patch('baelfire.application.application.Graph') as graph:
            with patch.object(ExampleTask, 'phase_build') as phase_build:
                with patch.object(ExampleTask, 'save_report') as save_report:
                    app.run_command_or_print_help(args)

                    save_report.assert_called_once_with()
                    phase_build.assert_called_once_with()
                    graph.assert_called_once_with(save_report.return_value)
                    graph.return_value.render.assert_called_once_with()

    def test_graph_drawing(self, app, parse_args):
        args = MagicMock()
        args.task = False
        with patch('baelfire.application.application.Graph') as graph:
            app.run_command_or_print_help(args)

            graph.assert_called_once_with(args.graph_file)
            graph.return_value.render.assert_called_once_with()



from mock import MagicMock

from baelfire.dependencies.task import TaskRebuilded
from baelfire.dependencies.task import RunTask


class TestRunTask(object):

    def test_fail(self):
        """
        RunTask should run linked task, but should not affect dependency
        validating.
        """
        task = MagicMock()
        dependency = RunTask(task)

        assert dependency.should_build() is False


class TestTaskRebuilded(object):

    def test_fail(self):
        """
        TaskRebuilded should run linked task, but should not affect dependency
        validating.
        """
        task = MagicMock()
        dependency = TaskRebuilded(task)

        assert dependency.should_build() is task.phase_validation.return_value



from baelfire.dependencies.dependency import Dependency


class TaskRebuilded(Dependency):
    """
    Trigger build parent task if assigned task has rebuilded.
    """

    def __init__(self, task):
        super(TaskRebuilded, self).__init__()
        self.task = task

    def phase_init(self):
        super(TaskRebuilded, self).phase_init()
        self.task = self.parent.top_core.prepere_task(self.task)
        self.task.phase_init()
        self.myreport['task'] = self.task.name

    def phase_data(self):
        super(TaskRebuilded, self).phase_data()
        self.task.phase_data()

    def set_parent(self, parent):
        super(TaskRebuilded, self).set_parent(parent)
        self.task.set_parent(parent)

    def should_build(self):
        return self.task.phase_validation()

    def build(self):
        self.task.phase_dependencies_build()
        self.task.phase_build()


class RunTask(TaskRebuilded):
    """
    Run provided task before the parent.
    """

    def should_build(self):
        super(RunTask, self).should_build()
        return False



from mock import patch
from pytest import fixture
from pytest import raises
from pytest import yield_fixture

from baelfire.task.screen import AttachScreenTask
from baelfire.task.screen import ScreenCore
from baelfire.task.screen import ScreenTask


class ExampleScreenTask(ScreenTask):
    screen_name = 'halp'

    def create_dependecies(self):
        pass


class ExampleAttachScreenTask(AttachScreenTask):
    detached_task = ExampleScreenTask


class ScreenFixtures(object):

    @yield_fixture
    def mpopen(self, task):
        with patch.object(task, 'popen') as mock:
            yield mock

    @fixture
    def core(self):
        core = ScreenCore()
        core.init()
        core.phase_settings()
        return core


class TestScreenTask(ScreenFixtures):

    @fixture
    def task(self, core):
        return ExampleScreenTask(core)

    def test_screen_run_without_name(self, task, mpopen):
        task.screen_name = None
        assert task._screen_run(['arg'], kw='arg2') == mpopen.return_value
        mpopen.assert_called_once_with(
            ['/usr/bin/screen -dm arg'],
            kw='arg2')

    def test_screen_run_with_name(self, task, mpopen):
        assert task._screen_run(['arg'], kw='arg2') == mpopen.return_value
        mpopen.assert_called_once_with(
            ['/usr/bin/screen -dmS halp arg'],
            kw='arg2')


class TestAttachScreenTask(ScreenFixtures):

    @fixture
    def task(self, core, mrun_before):
        return ExampleAttachScreenTask(core)

    @yield_fixture
    def mrun_before(self):
        with patch.object(ExampleAttachScreenTask, 'run_before') as mock:
            yield mock

    def test_no_detached_task_configured(self):
        with raises(AssertionError):
            AttachScreenTask()

    def test_screen_attach(self, task, mpopen, mrun_before):
        task.run()
        mpopen.assert_called_once_with(
            ['/usr/bin/screen -r halp'])
        mrun_before.assert_called_once_with(task._detached_task)



from baelfire.application.commands.graph.visualizators.chooser import VisualizatorChooser


class DependencyVisualization(object):
    PREFIX = '\t\t'
    DEPENDENCY_TEMPLATE = '"%(url)s"[label="%(name)s",fillcolor=%(fill)s,shape=box,style=filled];\n'

    def __init__(self, url, report, parent):
        self.parent = parent
        self.url = url
        self.report = report

    def dependency(self):
        return {
            'url': self.full_url,
            'name': self.get_name(),
            'fill': self.get_fill(),
        }

    def render(self):
        return self.PREFIX + self.DEPENDENCY_TEMPLATE % self.dependency()

    @property
    def full_url(self):
        return self.url + ':' + str(self.report['index'])

    def get_fill(self):
        if self.report['should_build']:
            if self.report['success']:
                return 'yellow'
            else:
                return 'white'
        return 'white'

    def get_name(self):
        return self.url.split('.')[-1]


class HiddenVisualization(DependencyVisualization):

    def render(self):
        return ''


class FileVisualization(DependencyVisualization):

    def get_name(self):
        return super(FileVisualization, self).get_name() + r'\n' + self.report['filename']


class DependencyVisualizatorChooser(VisualizatorChooser):
    DEFAULT = DependencyVisualization

    def generate_visualizators(self):
        self.add_visualizator(
            'baelfire.dependencies.dependency.AlwaysTrue',
            HiddenVisualization,
        )
        self.add_visualizator(
            'baelfire.dependencies.task.TaskRebuilded',
            HiddenVisualization,
        ),
        self.add_visualizator(
            'baelfire.dependencies.task.RunTask',
            HiddenVisualization,
        )
        self.add_visualizator(
            'baelfire.dependencies.file.FileDoesNotExists',
            FileVisualization,
        )
        self.add_visualizator(
            'baelfire.dependencies.file.FileChanged',
            FileVisualization,
        )



class TaskMustHaveOutputFileError(Exception):

    def __init__(self, name):
        self.name = name

    def __str__(self):
        return 'Error: Taks must have output_file setted: %s' % (self.name)


class CouldNotCreateFileError(Exception):

    def __init__(self, filename):
        self.filename = filename

    def __str__(self):
        return 'Error: Could not create file %s' % (self.filename)


class TaskNotFoundError(Exception):

    def __init__(self, task):
        self.task = task

    def __str__(self):
        return 'Error: Task "%s" can not be found!' % (self.task)


class OnlyOneTaskInARowError(Exception):

    def __init__(self, task):
        self.task = task

    def __str__(self):
        return 'Error: Task "%s" can be run only once!' % (self.task)


class CommandError(Exception):

    def __init__(self, number, text=''):
        self.number = number
        self.text = text

    def __str__(self):
        return 'Error: Command error (%d): %s' % (self.number, self.text)


class CommandAborted(Exception):
    pass



from subprocess import Popen
from yaml import load

from baelfire.application.commands.graph.visualizators.task import TaskVisualizatorChooser


class Graph(object):

    class Config(object):
        dot_path = 'graph.dot'
        png_path = 'graph.png'
        task = TaskVisualizatorChooser

    def __init__(self, path):
        self.path = path

    def read_report(self):
        file = open(self.path, 'r')
        return load(file)

    def render(self):
        self.generate_dot_file()
        self.generate_png_file()

    def generate_dot_file(self):
        self.report = self.read_report()
        with open(self.Config.dot_path, 'w') as stream:
            stream.write('digraph {\n')
            for key, element in sorted(self.report.items()):
                if type(element) is dict:
                    visualization = self.Config.task().choose(key, element)
                    stream.write(visualization.render())
            stream.write('}\n')

    def generate_png_file(self):
        spp = Popen(
            [
                'dot -Tpng %s -o %s' % (
                    self.Config.dot_path,
                    self.Config.png_path,
                )
            ],
            shell=True,
        )
        spp.wait()



from io import BytesIO
from mock import MagicMock
from mock import patch
from pytest import mark
from pytest import yield_fixture
from subprocess import PIPE

from baelfire.dependencies.screen import ScreenIsNotRunning


class TestScreenIsNotRunning(object):
    DATA_NO_SCREENS = b'''No Sockets found in /run/screens/S-socek.

'''
    DATA_SCREEN = b'''There is a screen on:
        14709.pts-4.gringo      (Attached)
        18963.elo       (Detached)
2 Sockets in /run/screens/S-socek.
'''

    @yield_fixture
    def mlist_screens(self):
        with patch.object(ScreenIsNotRunning, '_list_screen') as mock:
            yield mock

    @yield_fixture
    def mpopen(self):
        with patch('baelfire.dependencies.screen.Popen') as mock:
            yield mock

    def test_get_screens_when_no_screens(self, mlist_screens):
        mlist_screens.return_value = BytesIO(self.DATA_NO_SCREENS)
        dep = ScreenIsNotRunning(None)
        assert list(dep.get_screens()) == []

    def test_get_screens(self, mlist_screens):
        mlist_screens.return_value = BytesIO(self.DATA_SCREEN)
        dep = ScreenIsNotRunning(None)
        assert list(dep.get_screens()) == [
            {'name': 'pts-4.gringo', 'pid': 14709, 'status': 'Attached'},
            {'name': 'elo', 'pid': 18963, 'status': 'Detached'}]

    @mark.parametrize(
        "name, parent_name, expected",
        [
            ('name', 'parent_name', 'name'),
            (None, 'parent_name', 'parent_name')
        ]
    )
    def test_screen_name(self, name, parent_name, expected):
        parent = MagicMock()
        parent.screen_name = parent_name
        dep = ScreenIsNotRunning(name)
        dep.set_parent(parent)

        assert dep.screen_name == expected

    @mark.parametrize(
        'screen_name, expected',
        [
            ('elo', False),
            ('else', True),
        ]
    )
    def test_should_build(self, mlist_screens, screen_name, expected):
        mlist_screens.return_value = BytesIO(self.DATA_SCREEN)
        dep = ScreenIsNotRunning(screen_name)
        assert dep.should_build() == expected

    def test_list_screen(self, mpopen):
        parent = MagicMock()
        parent.paths.get.return_value = '/usr/bin/screen'
        dep = ScreenIsNotRunning(None)
        dep.set_parent(parent)

        assert dep._list_screen() == mpopen.return_value.stdout
        mpopen.assert_called_once_with(
            ['/usr/bin/screen -list'],
            shell=True,
            stdout=PIPE)
        mpopen.return_value.wait.assert_called_once_with()



from tempfile import NamedTemporaryFile
from time import sleep

from mock import MagicMock

from baelfire.dependencies.file import FileChanged
from baelfire.dependencies.file import FileDependency
from baelfire.dependencies.file import FileDoesNotExists
from baelfire.dependencies.file import FileExists


class TestFileDependency(object):

    def test_source(self):
        """
        FileDependency should accept source name wich will be later used as a
        key in .paths.
        """
        dependency = FileDependency('myname')
        parent = MagicMock()
        parent.paths = {'myname': 'success'}
        dependency.set_parent(parent)

        assert dependency.path == 'success'

    def test_phase_data(self):
        """
        FileDependency should add filename in report.
        """
        dependency = FileDependency('myname')
        parent = MagicMock()
        parent.paths = {'myname': 'success'}
        parent.myreport = {'dependencies': []}
        dependency.set_parent(parent)

        dependency.phase_data()

        assert dependency.myreport == {
            'filename': 'success',
            'name': 'baelfire.dependencies.file.FileDependency',
        }


class TestFileChanged(object):

    def test_on_output_not_existing(self):
        """
        FileChanged should indicate to rebuild if output file does not exists.
        """
        name = NamedTemporaryFile().name
        parent = MagicMock()
        parent.output = name
        dependency = FileChanged('tmp')
        dependency.set_parent(parent)

        assert dependency.should_build() is True

    def test_on_source_not_existing(self):
        """
        FileChanged should indicate to rebuild if source file does not exists.
        """
        name = NamedTemporaryFile(delete=False).name
        parent = MagicMock()
        parent.output = name
        dependency = FileChanged(raw_path=NamedTemporaryFile().name)
        dependency.set_parent(parent)

        assert dependency.should_build() is True

    def test_on_output_younger(self):
        """
        FileChanged should indicate to rebuild if output file is younger then
        the source file.
        """
        with NamedTemporaryFile(delete=False) as destination:
            destination.close()
            sleep(0.01)
            with NamedTemporaryFile(delete=False) as source:
                source.close()
                parent = MagicMock()
                parent.output = destination.name
                parent.paths = {
                    'source': source.name,
                }
                dependency = FileChanged('source')
                dependency.set_parent(parent)
                # import ipdb ; ipdb.set_trace()
                assert dependency.should_build() is True

    def test_on_output_older(self):
        """
        FileChanged should indicate not to rebuild if output file is older then
        the source file.
        """
        with NamedTemporaryFile(delete=False) as source:
            source.close()
            with NamedTemporaryFile(delete=False) as destination:
                destination.close()
                parent = MagicMock()
                parent.output = destination.name
                parent.paths = {
                    'source': source.name,
                }
                dependency = FileChanged('source')
                dependency.set_parent(parent)

                assert dependency.should_build() is False


class TestFileDoesNotExists(object):

    def test_on_file_not_exists(self):
        """
        FileDoesNotExists should indicate to rebuild if file is not existing.
        """
        name = NamedTemporaryFile().name
        parent = MagicMock()
        parent.paths = {
            'source': name,
        }
        dependency = FileDoesNotExists('source')
        dependency.set_parent(parent)

        assert dependency.should_build() is True

    def test_on_file_exists(self):
        """
        FileDoesNotExists should indicate not to rebuild if file exists.
        """
        with NamedTemporaryFile(delete=False) as source:
            parent = MagicMock()
            parent.paths = {
                'source': source.name,
            }
            dependency = FileDoesNotExists('source')
            dependency.set_parent(parent)

            assert dependency.should_build() is False


class TestFileExists(object):

    def test_on_file_not_exists(self):
        """
        FileExists should indicate not to rebuild if file is not existing.
        """
        name = NamedTemporaryFile().name
        parent = MagicMock()
        parent.paths = {
            'source': name,
        }
        dependency = FileExists('source')
        dependency.set_parent(parent)

        assert dependency.should_build() is False

    def test_on_file_exists(self):
        """
        FileExists should indicate to rebuild if file exists.
        """
        with NamedTemporaryFile(delete=False) as source:
            parent = MagicMock()
            parent.paths = {
                'source': source.name,
            }
            dependency = FileExists('source')
            dependency.set_parent(parent)

            assert dependency.should_build() is True



class VisualizatorChooser(object):

    def __init__(self):
        self.visualizators = {}
        self.generate_visualizators()

    def add_visualizator(self, url, visualizator):
        self.visualizators[url] = visualizator

    def choose(self, url, report, *args, **kwargs):
        visualizator = self.visualizators.get(url, self.DEFAULT)
        return visualizator(url, report, *args, **kwargs)



from morfdict import Paths
from morfdict import StringDict


class TaskInheritance(object):

    def __init__(self, base_cls, child_cls, migration=None):
        self.base_cls = base_cls
        self.child_cls = child_cls
        self.migration = migration or self._default_migration

    def is_ready_to_migrate(self, task):
        return self.base_cls == task.__class__

    def migrate(self, task):
        return self.migration(task, self.child_cls)

    def _default_migration(self, task, child_cls):
        return child_cls()

    def is_doubled(self, base_cls):
        return self.base_cls == base_cls


class Core(object):

    def __init__(self):
        self.inheritance = []
        self.make_task_inheritance()

    def init(self):
        self.settings = StringDict()
        self.paths = Paths()
        self.report = {'last_index': 0}

    def phase_settings(self):
        self.paths.set('jinja_templates', '', is_root=True)

    def prepere_task(self, task):
        for inheritance in self.inheritance:
            if inheritance.is_ready_to_migrate(task):
                old_parent = task.parent
                task = inheritance.migrate(task)
                task.set_parent(old_parent)
                return task
        return task

    def make_task_inheritance(self):
        pass

    def add_task_inheritance(self, base_cls, child_cls, migration=None):
        self._remove_old_inheritance(base_cls)
        self.inheritance.append(
            TaskInheritance(base_cls, child_cls, migration))

    def _remove_old_inheritance(self, base_cls):
        for inheritance in list(self.inheritance):
            if inheritance.is_doubled(base_cls):
                self.inheritance.remove(inheritance)
                return



from mock import MagicMock

from baelfire.parrented import parrented


class TestParrented(object):

    @property
    def parented(self):
        def method(self):
            return self
        return method

    def test_parrented_method_without_parent(self):
        """
        parrented should return own method if no parrent set.
        """
        obj = MagicMock()
        obj.parent = None
        method = parrented(self.parented)

        assert method(obj) == obj

    def test_parrented_method_with_parent(self):
        """
        parrented should return parent method if parrent is set.
        """
        parent = MagicMock()
        obj = MagicMock()
        obj.parent = parent
        method = parrented(self.parented)

        assert method(obj) == parent.method.return_value
        parent.method.assert_called_once_with(obj)

    def test_parrented_method_property(self):
        """
        parrented should return parent property.
        """
        parent = MagicMock()
        parent.method = 15
        obj = MagicMock()
        obj.parent = parent
        method = parrented(self.parented)

        assert method(obj) == parent.method



from baelfire.core import Core
from baelfire.dependencies import AlwaysTrue
from baelfire.dependencies.task import RunTask
from baelfire.task import Task


class ExampleTask(Task):

    def create_dependecies(self):
        self.build_if(AlwaysTrue())

    def build(self):
        self.top_core.flags['ExampleTask'] = True


class FakeTask(Task):

    def create_dependecies(self):
        self.build_if(AlwaysTrue())

    def build(self):
        self.top_core.flags['FakeTask'] = True


class ExampleTaskInherited(Task):

    def create_dependecies(self):
        self.build_if(AlwaysTrue())

    def build(self):
        self.top_core.flags['ExampleTaskInherited'] = True


class ExampleTaskInheritedSecond(Task):

    def create_dependecies(self):
        self.build_if(AlwaysTrue())

    def build(self):
        self.top_core.flags['ExampleTaskInheritedSecond'] = True


class RootTask(Task):

    def create_dependecies(self):
        self.build_if(RunTask(ExampleTask()))
        self.build_if(RunTask(FakeTask()))
        self.build_if(AlwaysTrue())

    def build(self):
        if self.parent:
            self.parent.core.flags['RootTask'] = True
        else:
            self.core.flags['RootTask'] = True


class DeepInheritanceTask(Task):

    def create_dependecies(self):
        self.build_if(RunTask(RootTask()))
        self.build_if(AlwaysTrue())

    def build(self):
        self.top_core.flags['DeepInheritanceTask'] = True


class MyCore(Core):

    def __init__(self):
        super(MyCore, self).__init__()
        self.flags = dict()

    def phase_settings(self):
        self.settings['one'] = 'one1'
        self.settings['three'] = 'three1'
        self.settings['two'] = '-%(one)s-'
        self.settings['four'] = '+%(three)s+'


class InheritanceCore(Core):

    def __init__(self):
        super(InheritanceCore, self).__init__()
        self.flags = dict()

    def make_task_inheritance(self):
        self.add_task_inheritance(ExampleTask, ExampleTaskInherited)


class TestCore(object):

    def test_simple_task(self):
        core = MyCore()
        task = ExampleTask(core)
        task.run()

        assert task.settings['two'] == '-one1-'
        assert task.settings['four'] == '+three1+'
        assert core.flags.get('ExampleTask')
        assert not core.flags.get('ExampleTaskInherited')


class TestInheritance(object):

    def test_inheritance_core(self):
        core = InheritanceCore()
        task = RootTask(core)
        task.run()

        assert not core.flags.get('ExampleTask')
        assert core.flags.get('ExampleTaskInherited')
        assert core.flags.get('RootTask')

    def test_no_inheritance_core(self):
        core = MyCore()
        task = RootTask(core)
        task.run()

        assert core.flags.get('ExampleTask')
        assert not core.flags.get('ExampleTaskInherited')
        assert core.flags.get('RootTask')

    def test_deep_inheritance(self):
        core = InheritanceCore()
        task = DeepInheritanceTask(core)
        task.run()

        assert not core.flags.get('ExampleTask')
        assert core.flags.get('ExampleTaskInherited')
        assert core.flags.get('RootTask')
        assert core.flags.get('DeepInheritanceTask')

    def test_double_inheritance(self):
        core = InheritanceCore()
        core.add_task_inheritance(ExampleTask, ExampleTaskInheritedSecond)
        core.add_task_inheritance(FakeTask, ExampleTaskInheritedSecond)
        task = DeepInheritanceTask(core)
        task.run()

        assert not core.flags.get('ExampleTask')
        assert not core.flags.get('ExampleTaskInherited')
        assert core.flags.get('ExampleTaskInheritedSecond')
        assert core.flags.get('RootTask')
        assert core.flags.get('DeepInheritanceTask')



from baelfire.dependencies.dependency import AlwaysTrue
from baelfire.dependencies.dependency import Dependency
from baelfire.dependencies.file import FileChanged
from baelfire.dependencies.file import FileDependency
from baelfire.dependencies.file import FileDoesNotExists
from baelfire.dependencies.file import FileExists
from baelfire.dependencies.pid import PidIsNotRunning
from baelfire.dependencies.pid import PidIsRunning
from baelfire.dependencies.task import RunTask
from baelfire.dependencies.task import TaskRebuilded


__all__ = [
    'AlwaysTrue',
    'Dependency',
    'FileChanged',
    'FileDependency',
    'FileDoesNotExists',
    'FileExists',
    'PidIsNotRunning',
    'PidIsRunning',
    'RunTask',
    'TaskRebuilded',
]



from mock import patch
from os import path
from pytest import fixture
from tempfile import NamedTemporaryFile

from baelfire.application.commands.graph.graph import Graph


def open_relative_file(filename):
    base = path.dirname(__file__)
    filename = path.join(base, filename)
    return open(filename, 'r')

EXAMPLE_REPORT = {
    '__main__.MyElo': {
        'aborted': False,
        'dependencies': [
            {
                'builded': True,
                'filename': '/tmp/elo',
                'index': 0,
                'phase_validation': True,
                'should_build': False,
                'success': True,
                'name': 'baelfire.dependencies.file.FileDoesNotExists',
            },
            {
                'builded': True,
                'index': 1,
                'phase_validation': True,
                'should_build': False,
                'success': True,
                'task': '__main__.MySecondElo',
                'name': 'baelfire.dependencies.task.TaskRebuilded',
            },
            {
                'builded': True,
                'filename': '/tmp/elo',
                'index': 2,
                'phase_validation': True,
                'should_build': True,
                'success': True,
                'name': 'baelfire.dependencies.file.FileChanged',
            },
            {
                'builded': True,
                'filename': '/tmp/elo',
                'index': 3,
                'phase_validation': True,
                'should_build': True,
                'success': False,
                'name': 'baelfire.dependencies.file.FileSomething',
            },
            {
                'builded': True,
                'filename': '/tmp/elo',
                'index': 4,
                'phase_validation': True,
                'should_build': True,
                'success': True,
                'name': 'baelfire.dependencies.dependency.AlwaysTrue',
            },
        ],
        'needtorun': False,
        'runned': False,
        'signal': None,
        'success': False},
    '__main__.MySecondElo': {
        'aborted': False,
        'dependencies': [],
        'needtorun': False,
        'runned': False,
        'signal': None,
        'success': False,
    },
    '__main__.Something': {
        'aborted': False,
        'dependencies': [],
        'needtorun': True,
        'runned': False,
        'signal': None,
        'success': False,
    },
    '__main__.Something2': {
        'aborted': False,
        'dependencies': [],
        'needtorun': True,
        'runned': True,
        'signal': None,
        'success': True,
    },
    'last_index': 2
}

EXPECTED_DOT_FILE = open_relative_file('expected_dot_file.txt').read()


class TestGraph(object):

    @fixture
    def path(self):
        return '/tmp/my/report.yaml'

    @fixture
    def obj(self, path):
        return Graph(path)

    def test_generate_png_file(self, obj):
        with patch('baelfire.application.commands.graph.graph.Popen') as popen:
            obj.generate_png_file()

            popen.assert_called_once_with(
                [
                    'dot -Tpng graph.dot -o graph.png',
                ],
                shell=True,
            )
            popen.return_value.wait.assert_called_once_with()

    def test_read_report(self, obj, path):
        open_patcher = patch('baelfire.application.commands.graph.graph.open')
        load_patcher = patch('baelfire.application.commands.graph.graph.load')
        with open_patcher as mopen:
            with load_patcher as mload:
                obj.read_report()

                mopen.assert_called_once_with(path, 'r')
                mload.assert_called_once_with(mopen.return_value)

    def test_render(self, obj):
        with patch.object(obj, 'generate_dot_file') as generate_dot_file:
            with patch.object(obj, 'generate_png_file') as generate_png_file:

                obj.render()

                generate_dot_file.assert_called_once_with()
                generate_png_file.assert_called_once_with()

    def test_with_fake_report(self, obj):
        patcher_read_report = patch.object(obj, 'read_report')
        with patcher_read_report as read_report:
            read_report.return_value = EXAMPLE_REPORT

            obj.Config.dot_path = NamedTemporaryFile().name
            obj.generate_dot_file()

            data = open(obj.Config.dot_path, 'r').read()
            assert EXPECTED_DOT_FILE == data.replace('\t', '    ')



from mock import MagicMock
from signal import SIGKILL
from subprocess import Popen
from tempfile import NamedTemporaryFile

from baelfire.dependencies.pid import PidIsNotRunning
from baelfire.dependencies.pid import PidIsRunning


class TestPidIsRunning(object):

    def test_raw_pid(self):
        spp = Popen(['sleep 10'], shell=True)

        is_running = PidIsRunning(spp.pid)
        is_not_running = PidIsNotRunning(spp.pid)

        try:
            assert is_running.should_build() is True
            assert is_not_running.should_build() is False
        finally:
            spp.send_signal(SIGKILL)

    def test_raw_path(self):
        spp = Popen(['sleep 10'], shell=True)
        with NamedTemporaryFile(mode='w', delete=False) as data:
            data.write(str(spp.pid))

            is_running = PidIsRunning(pid_file_path=data.name)
            is_not_running = PidIsNotRunning(pid_file_path=data.name)

        try:
            assert is_running.should_build() is True
            assert is_not_running.should_build() is False
        finally:
            spp.send_signal(SIGKILL)

    def test_path_name(self):
        spp = Popen(['sleep 10'], shell=True)
        with NamedTemporaryFile(mode='w', delete=False) as data:
            data.write(str(spp.pid))
            parent = MagicMock()
            parent.paths = {'pid': data.name}
        is_running = PidIsRunning(pid_file_name='pid')
        is_running.set_parent(parent)
        is_not_running = PidIsNotRunning(pid_file_name='pid')
        is_not_running.set_parent(parent)

        try:
            assert is_running.should_build() is True
            assert is_not_running.should_build() is False
        finally:
            spp.send_signal(SIGKILL)

    def test_pid_not_exists(self):
        is_running = PidIsRunning(1231231213)
        is_not_running = PidIsNotRunning(1231231213)

        assert is_running.should_build() is False
        assert is_not_running.should_build() is True



from signal import SIGALRM

from mock import MagicMock
from mock import call
from mock import patch
from pytest import fixture
from pytest import raises
from pytest import yield_fixture

from baelfire.error import CommandAborted
from baelfire.error import CommandError
from baelfire.task.process import SubprocessTask


class ExampleSubproccessTask(SubprocessTask):

    def create_dependecies(self):
        pass


class TestSubproccessTask(object):

    @fixture
    def task(self):
        return ExampleSubproccessTask()

    @yield_fixture
    def signal(self):
        with patch('baelfire.task.process.signal') as mock:
            yield mock

    @yield_fixture
    def popen(self):
        with patch('baelfire.task.process.Popen') as mock:
            yield mock

    @yield_fixture
    def send_signal(self, task):
        with patch.object(task, 'send_signal') as mock:
            yield mock

    @yield_fixture
    def task_popen(self, task):
        with patch.object(task, '_popen') as mock:
            yield mock

    @yield_fixture
    def init_signals(self, task):
        with patch.object(task, '_init_signals') as mock:
            yield mock

    def test_phase_init(self, task):
        """
        .phase_init should add .spp with None and report['aborted'] = False
        """
        task.phase_init()

        assert task.myreport['aborted'] is False
        assert task.myreport['signal'] is None

    def test_init_signals(self, task, signal):
        task._init_signals()

        signal.signal.assert_has_calls([
            call(signal.SIGABRT, task.on_signal),
            call(signal.SIGFPE, task.on_signal),
            call(signal.SIGILL, task.on_signal),
            call(signal.SIGINT, task.on_signal),
            call(signal.SIGSEGV, task.on_signal),
            call(signal.SIGTERM, task.on_signal),
        ])

    def test_send_signal_on_poll_something(self, task):
        """
        .send_signal do nothing, if spp.poll() is empty.
        spp.poll() returning something means the subprocess has ended
        """
        task.spp = MagicMock()

        task.send_signal('sig')

        task.spp.poll.assert_called_once_with()
        assert not task.spp.send_signal.called

    def test_send_signal_normal(self, task):
        """
        .send_signal should send signal if poll returned something
        """
        task.spp = MagicMock()
        task.spp.poll.return_value = None

        task.send_signal('sig')

        task.spp.poll.assert_called_once_with()
        task.spp.send_signal.assert_called_once_with('sig')

    def test_send_signal_on_error(self, task):
        """
        .send_signal should do nothing on raised OSError
        OSError means that subprocess returned error
        """
        task.spp = MagicMock()
        task.spp.poll.side_effect = OSError

        task.send_signal('sig')

        task.spp.poll.assert_called_once_with()
        assert not task.spp.send_signal.called

    def test_on_signal(self, task, send_signal):
        """
        .on_signal should update report and send signal to process when
        process exists
        """
        task.phase_init()
        task.spp = MagicMock()

        task.on_signal(SIGALRM, None)

        send_signal.assert_called_once_with(SIGALRM)
        assert task.myreport['aborted'] is True
        assert task.myreport['signal'] is SIGALRM

    def test_on_signal_with_no_spp(self, task, send_signal):
        """
        .on_signal should update report. This is sanity check.
        """
        task.phase_init()
        task.spp = None

        task.on_signal(SIGALRM, None)

        assert task.myreport['aborted'] is True
        assert task.myreport['signal'] is SIGALRM

    def test_popen(self, task, task_popen, init_signals):
        task.phase_init()
        task.spp = MagicMock()
        task.spp.returncode = 0
        task.popen('arg', kw='arg')

        init_signals.assert_called_once_with()
        task_popen.assert_called_once_with('arg', kw='arg', shell=True)

    def test_popen_on_error(self, task, task_popen, init_signals):
        task.phase_init()
        task.spp = MagicMock()
        task.myreport['aborted'] = True

        with raises(CommandAborted):
            task.popen()

        init_signals.assert_called_once_with()
        task_popen.assert_called_once_with(shell=True)

    def test_popen_on_command_error(self, task, task_popen, init_signals):
        task.phase_init()
        task.spp = MagicMock()

        with raises(CommandError):
            task.popen()

        init_signals.assert_called_once_with()
        task_popen.assert_called_once_with(shell=True)

    def test_system_popen(self, task, popen):
        task._popen('arg', kw='arg')

        popen.assert_called_once_with('arg', kw='arg')
        assert task.spp == popen.return_value
        task.spp.wait.assert_called_once_with()



from baelfire import error


class TestErrors(object):

    def test_task_must_have_output_file_error(self):
        """
        Should create TaskMustHaveOutputFileError error with name.
        """
        er = error.TaskMustHaveOutputFileError('name')
        assert str(er) == "Error: Taks must have output_file setted: name"

    def test_could_not_create_file_error(self):
        """
        Should create CouldNotCreateFileError error with filename.
        """
        er = error.CouldNotCreateFileError('filename')
        assert str(er) == 'Error: Could not create file filename'

    def test_task_not_found_error(self):
        """
        Should create TaskNotFoundError error with task name.
        """
        er = error.TaskNotFoundError('task_me')
        assert str(er) == 'Error: Task "task_me" can not be found!'

    def test_only_one_task_in_a_row_error_error(self):
        """
        Should create OnlyOneTaskInARowError error with task name.
        """
        er = error.OnlyOneTaskInARowError('task_me')
        assert str(er) == 'Error: Task "task_me" can be run only once!'

    def test_command_error(self):
        er = error.CommandError(11, 'text')
        assert str(er) == 'Error: Command error (11): text'



from logging import getLogger
from yaml import dump

from baelfire.core import Core
from baelfire.dependencies import RunTask
from baelfire.parrented import parrented


class Task(object):
    """
    Normal task.
    """

    @property
    def name(self):
        """
        Return name of this task.
        """
        return self.__module__ + "." + self.__class__.__name__

    @property
    @parrented
    def settings(self):
        return self.core.settings

    @property
    @parrented
    def paths(self):
        return self.core.paths

    @property
    @parrented
    def report(self):
        return self.core.report

    @property
    def myreport(self):
        if self.name not in self.report:
            self.report[self.name] = {}
        return self.report[self.name]

    @property
    def logger(self):
        return getLogger(self.name)

    def __init__(self, core=None):
        self._dependencies = []
        self.parent = None

        self._args = {}
        self.core = core or Core()
        self.create_dependecies()

    def run(self):
        """
        Check all dependencies and rebuild if needed.
        """
        self._step_validation()
        self._step_build()

    def _step_validation(self):
        self.phase_init()
        self.phase_settings()
        self.phase_data()
        self.phase_validation()

    def _step_build(self):
        self.phase_dependencies_build()
        self.phase_build()

    def phase_init(self):
        self.core.init()

        self.paths.set('report', '.baelfire.report')

        self.myreport['runned'] = False
        self.myreport['needtorun'] = False
        self.myreport['success'] = False
        self.myreport['dependencies'] = []

        for dependency in self._dependencies:
            self.myreport['dependencies'].append(dependency.myreport)
            dependency.phase_init()

    def phase_settings(self):
        self.core.phase_settings()

    def phase_data(self):
        for dependency in self._dependencies:
            dependency.phase_data()

    def phase_validation(self):
        for dependency in self._dependencies:
            result = dependency.phase_validation()
            self.myreport['needtorun'] |= result
        self.logger.debug('Need to run: %s', self.myreport['needtorun'])
        return self.myreport['needtorun']

    def phase_dependencies_build(self):
        for dependency in self._dependencies:
            dependency.phase_build()

    def phase_build(self):
        if self.myreport['needtorun'] and not self.myreport['runned']:
            self.logger.info('Running')
            self.myreport['runned'] = True
            try:
                self.build()
                self.myreport['success'] = True
            except Exception as error:
                self.logger.error(str(error))
                raise

    def build_if(self, dependency):
        """
        Add dependency to a task.
        """
        self._dependencies.append(dependency)
        dependency.set_parent(self)

    def run_before(self, task):
        """
        Run task before self.
        """
        dependency = RunTask(task)
        self._dependencies.append(dependency)
        dependency.set_parent(self)

    def set_parent(self, parent):
        self.parent = parent

    def get_index(self):
        index = self.report['last_index']
        self.report['last_index'] = index + 1
        return index

    def save_report(self):
        """
        Save report file after task run.
        """
        with open(self.paths.get('report'), 'w') as file:
            dump(self.report, file, default_flow_style=False)
        return self.paths.get('report')

    @property
    def top_core(self):
        if self.parent:
            return self.parent.top_core
        else:
            return self.core



from baelfire.application.commands.graph.visualizators.chooser import VisualizatorChooser


class LinkVisualization(object):
    PREFIX = '\t\t\t'
    LINK_TEMPLATE = '"%(right_url)s" -> "%(left_url)s";\n'
    HIDDEN_RIGHTS = [
        'baelfire.dependencies.dependency.AlwaysTrue',
    ]

    def __init__(self, left_url, left, right_url, right):
        self.left_url = left_url
        self.left = left
        self.right_url = right_url
        self.right = right

    def link(self):
        task = self.right.get('task')
        if task:
            right_url = task
        else:
            right_url = self.right_url

        return {
            'left_url': self.left_url,
            'right_url': right_url,
        }

    def render(self):
        if self.right['name'] in self.HIDDEN_RIGHTS:
            return ''
        return self.PREFIX + (self.LINK_TEMPLATE % self.link())


class LinkVisualizatorChooser(VisualizatorChooser):
    DEFAULT = LinkVisualization

    def generate_visualizators(self):
        pass



import logging
import os
import sys

from argparse import ArgumentParser
from importlib import import_module
from logging import getLogger

from baelfire.application.commands.graph.graph import Graph
from baelfire.error import TaskNotFoundError

log = getLogger(__name__)


class Application(object):

    def create_parser(self):
        self.parser = ArgumentParser()
        self._add_task_group()
        self._add_other_group()
        self._add_logging_group()

    def _add_task_group(self):
        tasks = self.parser.add_argument_group(
            'Tasks',
            'Tasks related options',
        )

        tasks.add_argument(
            '-t',
            '--task',
            dest='task',
            help='Run this task.',
        )
        tasks.add_argument(
            '-g',
            '--graph',
            dest='graph',
            help='Draw task dependency graph.',
            action="store_true",
        )

    def _add_other_group(self):
        other = self.parser.add_argument_group(
            'Other',
            'Other useful options',
        )
        other.add_argument(
            '-r',
            '--graph-file',
            dest='graph_file',
            help='Draw graph from report file.',
        )

    def _add_logging_group(self):
        log = self.parser.add_argument_group(
            'Logging:',
            'Logging related options.',
        )

        log.add_argument(
            '-l',
            '--log-level',
            dest='log_level',
            help='Log level',
            default='info',
            choices=[
                'debug',
                'info',
                'warning',
                'error',
                'critical',
            ]
        )

    def configure_logging(self, args):
        level = getattr(logging, args.log_level.upper())
        format = ' * %(levelname)s %(name)s: %(message)s *'
        logging.basicConfig(level=level, format=format)

    def run_command_or_print_help(self, args):
        if args.task:
            task = self.import_task(args.task)()
            try:
                try:
                    task.run()
                finally:
                    report_path = task.save_report()
            except:
                log.error('Error in %s' % (report_path,))
                raise
            if args.graph:
                Graph(report_path).render()
        elif args.graph_file:
            Graph(args.graph_file).render()
        else:
            self.parser.print_help()

    def import_task(self, package_url):
        """Returns task from url 'some.url:class'"""
        sys.path.append(os.getcwd())
        url = package_url.split(':')
        module = import_module(url[0])
        try:
            return getattr(module, url[1])
        except AttributeError:
            raise TaskNotFoundError(package_url)

    def run(self):
        self.create_parser()
        args = self.parser.parse_args()
        self.configure_logging(args)
        self.run_command_or_print_help(args)


def run():
    Application().run()



from jinja2 import Environment
from jinja2 import FileSystemLoader

from baelfire.dependencies import FileChanged
from baelfire.dependencies import FileDoesNotExists
from baelfire.task import FileTask


class BaseTemplateTask(FileTask):

    @property
    def source(self):
        return self.paths.get(self.source_name)

    def jinja(self):
        """Jinja2 environment generator."""
        return Environment(
            loader=self.get_jinja2_loader(),
            keep_trailing_newline=True,
        )

    def get_jinja2_loader(self):
        return FileSystemLoader(self.paths.get('jinja_templates'))

    def build(self):
        template = self.get_template()
        stream = template.stream(**self.generate_context())
        stream.dump(self.output)

    def generate_context(self):
        """Generates data which will be used in the template."""
        context = {}
        context['settings'] = self.settings
        context['paths'] = self.paths
        return context

    def get_template(self):
        return self.jinja().get_template(self.source)


class TemplateTask(BaseTemplateTask):
    """
    Task which generate file from template.
    """

    def create_dependecies(self):
        super(TemplateTask, self).create_dependecies()
        self.build_if(FileChanged(self.source_name))


class FirstTemplateTask(BaseTemplateTask):
    """
    Task which generate file from template but only if output does not exists.
    """

    def create_dependecies(self):
        super(FirstTemplateTask, self).create_dependecies()
        self.build_if(FileDoesNotExists(self.source_name))



# -*- encoding: utf-8 -*-
from setuptools import setup, find_packages

install_requires = [
    'morfdict>=0.4.1',
    'pytest',
    'jinja2',
    'mock',
    'pytest-cov',
    'pyyaml',
]

if __name__ == '__main__':
    setup(
        name='baelfire',
        version='0.5.0',
        description=("Tool for build automatization and a task runner"),
        author='Dominik "Socek" Długajczyk',
        author_email='msocek@gmail.com',
        packages=find_packages(),
        install_requires=install_requires,
        license='Apache License 2.0',
        package_data={
            '': ['baelfire/application/commands/graph/tests/*.txt'],
        },
        entry_points={
            'console_scripts': [
                'bael=baelfire.application.application:run',
            ]
        },
    )



from yaml import dump
from yaml import load
from os.path import exists


class FileDict(dict):

    def __init__(self, path):
        self.path = path

    def load(self, ignore_errors=False):
        self.clear()
        if ignore_errors and not self.exists():
            return self
        with open(self.path, 'r') as file:
            self.update(load(file))
        return self

    def exists(self):
        return exists(self.path)

    def save(self):
        with open(self.path, 'w') as file:
            dump(dict(self), file, default_flow_style=False)

    def ensure_key_exists(self, key, description=''):
        if key not in self:
            self.read_from_stdin(key, description)

    def read_from_stdin(self, key, description=''):
        if description:
            description += ': '
        self[key] = input(description)



from tempfile import NamedTemporaryFile

from baelfire.core import Core
from baelfire.task.template import FirstTemplateTask
from baelfire.task.template import TemplateTask


class ExampleCore(Core):
    source_name = 'source'
    output_name = 'output'

    def __init__(self):
        super(ExampleCore, self).__init__()
        self.source_file = NamedTemporaryFile(delete=False)
        self.source_file.write(
            b"{{paths.get('source')}}|{{paths.get('output')}}",
        )
        self.source_file.close()

    def phase_settings(self):
        super(ExampleCore, self).phase_settings()
        self.paths.set('jinja_templates', '', is_root=True)
        self.paths.set(self.source_name, self.source_file.name)
        self.paths.set(self.output_name, NamedTemporaryFile().name)


class ExampleTemplateTask(TemplateTask):
    source_name = 'source'
    output_name = 'output'


class ExampleFirstTemplateTask(FirstTemplateTask):
    source_name = 'source'
    output_name = 'output'


class TestTemplateTask(object):

    def test_normal(self):
        """
        Source should be a template to output file.
        """
        task = ExampleTemplateTask(ExampleCore())
        task.run()

        data = open(task.output, 'r').read()
        assert data == task.source + "|" + task.output

    def test_rebuild_on_template_change(self):
        """
        Task should be rebuild on template change.
        """
        core = ExampleCore()
        task = ExampleTemplateTask(core)
        task.run()

        with open(core.source_file.name, 'w') as data:
            data.write('xxx')

        task.run()

        data = open(task.output, 'r').read()
        assert data == 'xxx'


class TestFirstTemplateTask(object):

    def test_normal(self):
        """
        Source should be a template to output file.
        """
        task = ExampleFirstTemplateTask(ExampleCore())
        task.run()

        data = open(task.output, 'r').read()
        assert data == task.source + "|" + task.output

        with open(task.output, 'w') as fp:
            fp.write('new')

        task = ExampleFirstTemplateTask(ExampleCore())
        task.run()

        data = open(task.output, 'r').read()
        assert data == task.source + "|" + task.output
